/*
 *
 * Copyright (C) 2000  Robert Olsson, Jens Ls 
 * Swedish University of Agricultural Sciences
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#define MRT_HASH_SIZE 4096

struct mrt {
  u_int32_t flags;
  struct msrc *ms;

  /*u_int32_t group; */
  struct mgrp *mg;
  
  struct thread *t_entry_timeout;
  list   oif_list;
  time_t uptime;
  time_t exptime;

  struct thread *t_register_stop;
  struct thread *t_assert;

  /* Assert handling */
  struct {
    u_int32_t         preference; /* PIM */
    u_int32_t         metric;     /* PIM */
    u_int32_t         rpf_neighbor; /* PIM */
  } asserted;

  /* Latest counters from SG entry */

  struct {
    u_int32_t packets;
    u_int32_t bytes;
    u_int32_t old_packets;
    u_int32_t old_bytes;
    u_int32_t wrong_if;
    time_t    time;
    time_t    old_time;
  } stats ;
};


#define MSRC_HASH_SIZE 1024

#define MS_PEROIDIC_TIMER 60
struct thread *t_ms_timeout;


struct msrc {
  u_int32_t  use_count;
  struct interface *ifp;
  u_int32_t  rpf_neighbor; /* PIM */
  u_int32_t  src; 
  u_int32_t  preference;  /* PIM */
  u_int32_t  metric;      /* PIM */
  time_t     uptime;
};

int entry_timeout_thread(struct thread *t_mr_timeout);
struct mrt *mr_lookup(u_int32_t src, u_int32_t group );
struct mrt *mr_add(struct msrc *ms, struct mgrp *mg );
int mr_remove(struct mrt *mr);
struct oif *mr_oif_lookup(struct mrt *mr, struct interface *ifp);
void mr_add_oif(struct mrt *mr, struct interface *ifp);
int ms_timeout_thread(struct thread *t_ms_timeout);

void mr_timer_update(struct mrt *mr, time_t newtime);

struct msrc *ms_lookup(u_int32_t S);
struct msrc *ms_add(u_int32_t S);
int ms_remove(u_int32_t S );
int ms_list(struct vty *vty);

#define MGRP_HASH_SIZE 1024

struct oif {
  struct interface    *ifp;

  struct mrt          *mr;
  struct mgrp         *mg;
  u_int32_t           flags;
  time_t              uptime;             /* IGMP & PIM */
  
  struct {
    struct thread       *t_igmp_timeout;  /* IGMP */
    struct thread       *t_send_group_specific_query; /* IGMP */    
    u_int32_t           last_reporter;      /* IGMP */
    time_t              exptime_igmp;       /* IGMP */
    u_int32_t           leave_count;             /* IGMP */
    time_t              leave_request_time; /* IGMP When we got the leave request */
  } igmp;
  
  struct {
    struct thread       *t_oif_timeout;   /* PIM */
    time_t              exptime;            /* PIM */
    time_t              oif_deletion_delay;  /* PIM */
  } pim;  
};


#define MRT_LOCAL     (1<<0)
#define MRT_CONNECTED (1<<1)
#define MRT_PRUNED    (1<<2)
#define MRT_WC        (1<<3)
#define MRT_RPT       (1<<5)
#define MRT_SPT       (1<<6)
#define MRT_SPARSE    (1<<7)
#define MRT_DENSE     (1<<8)
#define MRT_REGISTER  (1<<9)
#define MRT_IGMP_IN_USE (1<<10)
#define MRT_PIM_IN_USE (1<<11)
#define MRT_REGISTER_STOP (1<<12)
#define MRT_BLACKHOLE (1<<13)
#define MRT_ASSERTED  (1<<14)
#define MRT_ALL       ((1<<15)-1)

struct mgrp {
  u_int32_t         group;
  u_int32_t         flags;
  
  struct {
    struct msrc *rp; 
    time_t            uptime;      /* PIM */
    time_t            exptime;     /* PIM */
    struct thread *t_entry_timeout; /* PIM */
    struct thread *t_assert;        /* PIM */

    /* Assert handling */
    struct {
      u_int32_t         preference; /* PIM */
      u_int32_t         metric;     /* PIM */
      u_int32_t         rpf_neighbor; /* PIM */
    } asserted;
  } pim;
  
  struct {
    time_t            last_heard;  /* IGMP */
  } igmp;
  
  list              oif_list;
  list              mrt_list;
};

void tables_init( void );

struct mgrp *mg_lookup(u_int32_t G );
struct mgrp *mg_add(u_int32_t G );
int mg_remove( struct mgrp *mg );
int mgrp_list(struct vty *vty);

int mg_oif_list_add(struct mgrp *mg, struct oif *mi);
int mr_oif_list_add(struct mrt *mr, struct oif *mi);
int mg_oif_list_delete (struct oif *mi);
int mr_oif_list_delete (struct oif *mi);
int mr_oif_delete(struct mrt *mr, struct interface *ifp);

void mgrp_timer_update(struct mgrp *mg, time_t newtime);

vifbitmap_t mr_oifs(struct mrt *mr);
u_int32_t mg_iif( struct mgrp *mg);
u_int32_t mr_iif( struct mrt *mr);

void add_oif_to_mr(struct mgrp *mg, struct interface *ifp);
void copy_mg_oifs(struct mrt *mr, struct mgrp *mg);
void sync_mg_family(struct mgrp *mg);
void mr_entry_timer_update(struct mrt *mr, time_t newtime);
void mg_pim_timer_update(struct mgrp *mg, time_t newtime);


int mr_flag_oif(struct mrt *mr, struct interface *ifp, u_int32_t flag);
int mg_flag_oif(struct mgrp *mg, struct interface *ifp, u_int32_t flag);
int mr_mfc_add( struct mrt *mr );
int mg_mfc_add( struct mgrp *mg );

u_int32_t mg_rpf_neighbor( struct mgrp *mg );
u_int32_t mr_rpf_neighbor( struct mrt *mr );
